@@ -1,21 +1,17 @@
 /*
-    Copyright 2005-2014 Intel Corporation.  All Rights Reserved.
+    Copyright (c) 2005-2019 Intel Corporation
 
-    This file is part of Threading Building Blocks. Threading Building Blocks is free software;
-    you can redistribute it and/or modify it under the terms of the GNU General Public License
-    version 2  as  published  by  the  Free Software Foundation.  Threading Building Blocks is
-    distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
-    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-    See  the GNU General Public License for more details.   You should have received a copy of
-    the  GNU General Public License along with Threading Building Blocks; if not, write to the
-    Free Software Foundation, Inc.,  51 Franklin St,  Fifth Floor,  Boston,  MA 02110-1301 USA
+    Licensed under the Apache License, Version 2.0 (the "License");
+    you may not use this file except in compliance with the License.
+    You may obtain a copy of the License at
 
-    As a special exception,  you may use this file  as part of a free software library without
-    restriction.  Specifically,  if other files instantiate templates  or use macros or inline
-    functions from this file, or you compile this file and link it with other files to produce
-    an executable,  this file does not by itself cause the resulting executable to be covered
-    by the GNU General Public License. This exception does not however invalidate any other
-    reasons why the executable file might be covered by the GNU General Public License.
+        http://www.apache.org/licenses/LICENSE-2.0
+
+    Unless required by applicable law or agreed to in writing, software
+    distributed under the License is distributed on an "AS IS" BASIS,
+    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+    See the License for the specific language governing permissions and
+    limitations under the License.
 */
 
 #include "tbb/blocked_range3d.h"
@@ -28,7 +24,7 @@ template<typename Tag>
 class AbstractValueType {
     AbstractValueType() {}
     int value;
-public: 
+public:
     template<typename OtherTag>
     friend AbstractValueType<OtherTag> MakeAbstractValueType( int i );
 
@@ -70,22 +66,22 @@ static void SerialTest() {
     typedef AbstractValueType<RowTag> row_type;
     typedef AbstractValueType<ColTag> col_type;
     typedef tbb::blocked_range3d<page_type,row_type,col_type> range_type;
-    for( int pagex=-4; pagex<4; ++pagex ) {
-        for( int pagey=pagex; pagey<4; ++pagey ) {
-            page_type pagei = MakeAbstractValueType<PageTag>(pagex);
-            page_type pagej = MakeAbstractValueType<PageTag>(pagey);
-            for( int pageg=1; pageg<4; ++pageg ) {
-                for( int rowx=-4; rowx<4; ++rowx ) {
-                    for( int rowy=rowx; rowy<4; ++rowy ) {
-                        row_type rowi = MakeAbstractValueType<RowTag>(rowx);
-                        row_type rowj = MakeAbstractValueType<RowTag>(rowy);
-                        for( int rowg=1; rowg<4; ++rowg ) {
-                            for( int colx=-4; colx<4; ++colx ) {
-                                for( int coly=colx; coly<4; ++coly ) {
-                                    col_type coli = MakeAbstractValueType<ColTag>(colx);
-                                    col_type colj = MakeAbstractValueType<ColTag>(coly);
-                                    for( int colg=1; colg<4; ++colg ) {
-                                        range_type r( pagei, pagej, pageg, rowi, rowj, rowg, coli, colj, colg );
+    for( int page_x=-4; page_x<4; ++page_x ) {
+        for( int page_y=page_x; page_y<4; ++page_y ) {
+            page_type page_i = MakeAbstractValueType<PageTag>(page_x);
+            page_type page_j = MakeAbstractValueType<PageTag>(page_y);
+            for( int page_grain=1; page_grain<4; ++page_grain ) {
+                for( int row_x=-4; row_x<4; ++row_x ) {
+                    for( int row_y=row_x; row_y<4; ++row_y ) {
+                        row_type row_i = MakeAbstractValueType<RowTag>(row_x);
+                        row_type row_j = MakeAbstractValueType<RowTag>(row_y);
+                        for( int row_grain=1; row_grain<4; ++row_grain ) {
+                            for( int col_x=-4; col_x<4; ++col_x ) {
+                                for( int col_y=col_x; col_y<4; ++col_y ) {
+                                    col_type col_i = MakeAbstractValueType<ColTag>(col_x);
+                                    col_type col_j = MakeAbstractValueType<ColTag>(col_y);
+                                    for( int col_grain=1; col_grain<4; ++col_grain ) {
+                                        range_type r( page_i, page_j, page_grain, row_i, row_j, row_grain, col_i, col_j, col_grain );
                                         AssertSameType( r.is_divisible(), true );
 
                                         AssertSameType( r.empty(), true );
@@ -94,13 +90,13 @@ static void SerialTest() {
                                         AssertSameType( static_cast<range_type::row_range_type::const_iterator*>(0), static_cast<row_type*>(0) );
                                         AssertSameType( static_cast<range_type::col_range_type::const_iterator*>(0), static_cast<col_type*>(0) );
 
-                                        AssertSameType( r.pages(), tbb::blocked_range<page_type>( pagei, pagej, 1 ));
-                                        AssertSameType( r.rows(), tbb::blocked_range<row_type>( rowi, rowj, 1 ));
-                                        AssertSameType( r.cols(), tbb::blocked_range<col_type>( coli, colj, 1 ));
+                                        AssertSameType( r.pages(), tbb::blocked_range<page_type>( page_i, page_j, 1 ));
+                                        AssertSameType( r.rows(), tbb::blocked_range<row_type>( row_i, row_j, 1 ));
+                                        AssertSameType( r.cols(), tbb::blocked_range<col_type>( col_i, col_j, 1 ));
 
-                                        ASSERT( r.empty()==(pagex==pagey||rowx==rowy||colx==coly), NULL );
+                                        ASSERT( r.empty()==(page_x==page_y||row_x==row_y||col_x==col_y), NULL );
 
-                                        ASSERT( r.is_divisible()==(pagey-pagex>pageg||rowy-rowx>rowg||coly-colx>colg), NULL );
+                                        ASSERT( r.is_divisible()==(page_y-page_x>page_grain||row_y-row_x>row_grain||col_y-col_x>col_grain), NULL );
 
                                         if( r.is_divisible() ) {
                                             range_type r2(r,tbb::split());
@@ -161,19 +157,45 @@ void ParallelTest() {
                              Array[l][m][n] = 0;
                         }
                     }
-                }   
+                }
             }
         }
     }
 }
 
+#if __TBB_CPP17_DEDUCTION_GUIDES_PRESENT
+#include <vector>
+void TestDeductionGuides() {
+    std::vector<const unsigned long *> v;
+    std::vector<double> v2;
+    std::vector<std::vector<int>> v3;
+
+    // check blocked_range2d(PageValue, PageValue, size_t, RowValue, RowValue, size_t, ColValue, ColValue, size_t)
+    tbb::blocked_range3d r1(v.begin(), v.end(), 2, v2.begin(), v2.end(), 2, v3.begin(), v3.end(), 6);
+    static_assert(std::is_same<decltype(r1),
+        tbb::blocked_range3d<decltype(v)::iterator, decltype(v2)::iterator, decltype(v3)::iterator>>::value);
+
+    // check blocked_range2d(blocked_range3d &)
+    tbb::blocked_range3d r2(r1);
+    static_assert(std::is_same<decltype(r2), decltype(r1)>::value);
+
+    // check blocked_range2d(blocked_range3d &&)
+    tbb::blocked_range3d r3(std::move(r1));
+    static_assert(std::is_same<decltype(r2), decltype(r1)>::value);
+}
+#endif
+
 #include "tbb/task_scheduler_init.h"
 
 int TestMain () {
-    SerialTest(); 
+    SerialTest();
     for( int p=MinThread; p<=MaxThread; ++p ) {
         tbb::task_scheduler_init init(p);
         ParallelTest();
     }
+
+    #if __TBB_CPP17_DEDUCTION_GUIDES_PRESENT
+        TestDeductionGuides();
+    #endif
     return Harness::Done;
 }
